# This Python file uses the following encoding: utf-8
import logging

from rumgeneric import generic, PredicateBuilder
import prioritized_methods
import peak.rules.core as rules_core
from inspect import isclass

class SecurityDenial(Exception):
    pass
log = logging.getLogger(__name__)

nested_attr='''attr is not None and attr.find('.') >=0'''
class PredicateBuilder(PredicateBuilder):
    def not_nested_attr(self):
        self._conditions.append("""not(%s)""" % (nested_attr,))
class Denial(object):
    """
    An object which is returned by :meth:`Policy.has_permission` when access
    is denied.

    It evaluates to False and holds a message with the reason for the denial
    """
    def __init__(self, msg):
        self.msg = msg

    def __nonzero__(self):
        return False

    def __unicode__(self):
        return unicode(self.msg)

class MultiDenial(Denial):
    def __init__(self, denials):
        self.denials = denials

    def __unicode__(self):
        #TODO: fix internationalization
        #return unicode(policy._(' or ')).join(map(unicode, self.denials))
        return unicode(' or ').join(map(unicode, self.denials))
    
class NotProvided:
    """
    A placeholder for when no optional argument is provided.
    """

class Any(rules_core.Before):
    """
    A Custom method that has special chaining semantics:

    When several of these rules apply then they're all evaluated (with
    no ordering guarantees) and if any of them returns True then access is
    granted. Else a Denial with all denial messages is returned
    """
    def __call__(self, *args, **kw):
        results = []
        for sig, body in self.sorted():
            results.append(body(*args, **kw))
        return any(results) or MultiDenial(results)

# "around" methods always override "any" methods
rules_core.always_overrides(rules_core.Around, Any)
# "any" methods always override "when" methods
rules_core.always_overrides(Any, rules_core.Method)
rules_core.merge_by_default(Any)

add_predicate = Any.make_decorator('add_predicate')
 
def base(string):
    pos = string.find('.')
    return string[:pos]

def remote(string):
    pos = string.find('.')
    return string[pos+1:]

class Policy(object):
    """
    Controls access to resources. It follows the strict design, 
    that every access must be explicitely allowed by conditionally
    applied predicate.
    Predicates  are registered using :meth:`register` for certain kind
    of `obj`, `action`, `attr`. When several rules match
    a given access, then at least one rule must grant the access, by evaluating
    the predicate positively.
    The Policy has to enforced on controller level for granting access to an object.
    It also has be made sure in view factories, that only fields/attributes are displayed, where access
    is granted.
    """


    def check(self, obj, action, attr=None, user=NotProvided):
        """
        Checks if `user` allowed to perform `action` on `obj`? Optional `attr`
        is the name of an attribute from `obj` which if present means the
        query is on that attribute, else it's on the whole `obj`.

        If `user` is :class:`NotProvided` then the currently logged in user
        will be used.

        Raises a :exc:`SecurityDenial` if access is denied
        """
        allowed = self.has_permission(obj, action, attr, user)
        if not allowed:
            raise self.exception_class(unicode(allowed))

    def has_permission(self, obj, action, attr=None, user=NotProvided):
        """
        Is `user` allowed to perform `action` on `obj`? Optional `attr`
        is the name of an attribute from `obj` which if present means the
        query is on that attribute, else it's on the whole `obj`.

        If `user` is :class:`NotProvided` then the currently logged in user
        will be used.

        Returns True or a :class:`Denial` instance.
        """
        return Denial(self._(u"No permission declared"))

    has_permission.add_predicate = add_predicate.__get__(has_permission)
    


    
    def get_current_user(self):
        raise NotImplementedError
    



        
    @classmethod
    def register(cls, predicate, obj=None, action=None, attr=None, rule=None):
        """
        Registers a `predicate` callable to be evaluated when
        :meth:`has_permission` is called with `obj`, `action`, and `attr` as
        arguments. `predicate` should be a callable that accepts the same
        arguments as :meth:`has_permission` and return True/Denial. Unspecified
        arguments act as wildcards.
        """
        locals_ = {}
        builder = PredicateBuilder(locals_, rule)
        if obj is not None:
            builder.subclass_or_instance("obj", obj)
        if action is not None:
            builder.equals("action", action)
        if attr is not None:
            builder.equals_or_none("attr", attr)

        context = rules_core.ParseContext(predicate, Any.make, locals_, {})
        has_permission = cls.has_permission.im_func
        engine = rules_core.Dispatching(has_permission).engine
        rules = rules_core.rules_for(has_permission)
        rule = rules_core.parse_rule(engine, builder.predicate, context, cls)
        rules.add(rule)
    

    @generic
    def filter(self, resource, query, user=NotProvided):
        """
        Applies extra-filtering criteria to a :class:`rum.interfaces.IQuery` so only
        allowed objects are returned.
        """

    
    @rules_core.around(has_permission, "user is NotProvided")
    #right Policy?
    def __use_current_user(next_method, self, obj, action, attr, user):
        try:
            current_user = self.get_current_user()
        except Exception, e:
            log.warn(e)
            return Denial(
                self._("No request is available to fetch the current user from")
                )
        return self.has_permission(
            obj=obj, action=action, attr=attr, user=current_user
            )
    
    def fields_for_resource(self, obj):
        from rum import app
        return app.fields_for_resource(obj) or app.fields_for_resource(obj.__class__)
       
    @rules_core.around(has_permission, "user is not NotProvided and "+nested_attr)
    def __handle_nested_attrs(next_method, self, obj, action, attr, user):
        res1=self.has_permission(obj, action, base(attr), user)
        
        if not res1:
            return res1
        res2=self.has_permission(obj, 'join', base(attr), user)
        if not res2:
            return res2
        fields_for_resource= self.fields_for_resource(obj)
        for f in fields_for_resource:
            if f.name ==  base(attr) and hasattr(f, 'other'):
                res3=self.has_permission(f.other,
                    action,
                    remote(attr),
                    user)
                return res3
            
        return Denial('a related field with correct name has to be registered')
    
    
    @filter.around("user is NotProvided")
    def __filter_with_current_user(self, resource, query, user):
        from rum import app
        try:
            current_user = self.get_current_user()
        except:
            current_user = None
        return self.filter(resource, query, current_user)

    @classmethod
    def register_filter(cls, filter, resource=None, rule=None):
        locals_ = {}
        builder = PredicateBuilder(locals_, rule)
        if resource is not None:
            builder.subclass("resource", resource)

        maker = prioritized_methods.PrioritizedMethod.make
        context = rules_core.ParseContext(filter, maker, locals_, {})
        func = cls.filter.im_func
        engine = rules_core.Dispatching(func).engine
        rules = rules_core.rules_for(func)
        rule = rules_core.parse_rule(engine, builder.predicate, context, cls)
        rules.add(rule)



    @filter.when(prio=-1)
    def __no_filter(self, resource, query, user):
        return query
    
    def _(self, string):
        "dummy internationalization"
        return string

    exception_class = SecurityDenial

register = Policy.register


    # except AttributeError:
    #     return Denial('Access to remote resource %s needs to be granted' %(attr,))


# Policy.register(__related_fields_obj, rule='''
# isinstance(attr, basestring) and attr.find('.') >0
#     and attr.find('.') >0
# ''', nested=True)


# Some useful predicates

def not_anonymous(policy, obj, action,  attr, user):
    if user is not None:
        return True
    else:
        return Denial(policy._("Need to authenticate yourself to access this resource"))

def anyone(policy, obj, action,  attr, user):
    return True

filter = Policy.filter.im_func
has_permission = Policy.has_permission

# A dummy default policy that always grants access
class DummyPolicy(Policy): pass
DummyPolicy.register(anyone)
#



from setuptools import setup, find_packages
import sys, os

version = '0.0.10'

setup(name='rum-policy',
      version=version,
      description="",
      long_description="""\
""",
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='security,rum,peak.rules',
      author='Alberto Vaverde Gonzalez, Michael Brickenstein',
      author_email='rum-discuss@googlegroups.com',
      url='http://bitbucket.org/brickenstein/rum-policy',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          # -*- Extra requirements: -*-
          "rum-generic"
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



from unittest import TestCase
import rumpolicy as policy
from rum.fields import FieldFactory, Field, Relation

class TestingPolicy(policy.Policy):
    def fields_for_resource(self, obj):
        return [
            Field('hotness'),
            Relation('chili2', remote_name=None, other=Chili),
            Relation('chili', remote_name=None, other=Chili),
            Relation('cookie', remote_name=None, other=Cookie)
        ]

class Edible(object): pass
class Potato(Edible): pass
class Tomato(Edible): pass
class Chili(Edible): pass
class Cookie(Edible): pass



class User(object):
    def __init__(self, permissions):
        self.permissions = permissions

def all_permissions(*perms):
    def check(self, obj, action, attr, user):
        allowed = all(map(user.permissions.__contains__, perms))
        if not allowed:
            return policy.Denial("Need to have all: %r"%(perms,))
        return True
    return check

def any_permission(*perms):
    def check(self, obj, action, attr, user):
        allowed = any(map(user.permissions.__contains__, perms))
        if not allowed:
            return policy.Denial("Need to have at least one: %r"%(perms,))
        return True
    return check

# A permission on an action on an obj
TestingPolicy.register(
    predicate=any_permission("edit", "edit_potato"),
    obj=Potato,
    action="edit"
    )

TestingPolicy.register(
    predicate=policy.not_anonymous,
    obj=Cookie
)

# A permission on an action and any obj
TestingPolicy.register(
    predicate=any_permission("edit"),
    action="edit"
    )

# Overlapping rules
TestingPolicy.register(
    predicate=any_permission("new"),
    obj=Tomato,
    action="new"
    )
TestingPolicy.register(
    predicate=any_permission("new_tomato"),
    obj=Tomato,
    action="new"
    )

TestingPolicy.register(
    predicate=any_permission("cook"),
    obj=Chili,
    action="cook",
    attr="hotness"
)
TestingPolicy.register(
    predicate=any_permission("cook"),
    obj=Chili,
    action="cook",
    attr="cookie"
)

TestingPolicy.register(
    predicate=any_permission("cook"),
    obj=Cookie,
    action="cook",
    attr="sweetness"
)

TestingPolicy.register(
    predicate=any_permission('join'),
    obj=Tomato,
    action='join')
TestingPolicy.register(
    predicate=any_permission('cook'),
    obj=Tomato,
    action='cook', attr='chili')

TestingPolicy.register(
    predicate=any_permission("bake"),
    obj=Chili,
    action="bake",
    rule="attr!='hotness'"
)

    
class TestPolicy(TestCase):
    def makeOne(self):
        return TestingPolicy()

    def test_register_action_on_obj(self):
        policy = self.makeOne()
        allowed = policy.has_permission(
            Potato(), 'edit', user=User(['edit_potato'])
            )
        self.failUnless(allowed)

        allowed = policy.has_permission(Potato(), 'edit', user=User(['add']))
        self.failUnless(not allowed, allowed)

    def test_register_action(self):
        policy = self.makeOne()
        allowed = policy.has_permission(Potato(), 'edit', user=User(['edit']))
        self.failUnless(allowed)
        allowed = policy.has_permission(Tomato(), 'edit', user=User(['edit']))
        self.failUnless(allowed)
    
    def test_register_exclusion_rule(self):
        policy = self.makeOne()
        user = User(['bake'])
        allowed = policy.has_permission(Chili(), 'bake', user=user)
        self.failUnless(allowed)
        allowed = policy.has_permission(Chili(), 'bake', user=user, attr='hotness')
        self.failUnless(not allowed)
        allowed = policy.has_permission(Chili(), 'bake', user=user, attr='amount')
        self.failUnless(allowed)
    
    def test_not_anonymous(self):
        policy = self.makeOne()
        allowed=policy.has_permission(Cookie(), 'eat them all')
        self.failUnless(not allowed, allowed)
        allowed=policy.has_permission(Cookie(), 'eat them all', user=User([]))
        self.failUnless(allowed)

    def test_register_overlapping_rules(self):
        # When two rules with the same preference exist then if any one of
        # them grants permission then access is granted
        policy = self.makeOne()
        allowed = policy.has_permission(Tomato(), 'new', user=User(['new']))
        self.failUnless(allowed)
        allowed = policy.has_permission(Tomato(), 'new', user=User(['new_tomato']))
        self.failUnless(allowed)
        allowed = policy.has_permission(Tomato(), 'new', user=User(['foo']))
        self.failUnless(not allowed)

    def test_deny_when_no_rules_apply(self):
        policy = self.makeOne()
        allowed = policy.has_permission(Edible(), 'new', user=User(['new']))
        self.failUnless(not allowed)

        
    def make_nested_tomato(self):
        tomato=Tomato()
        chili=Chili()
        cookie=Cookie()
        tomato.cookie=cookie
        tomato.chili=chili
        chili.cookie=cookie
        tomato.chili2=chili
        return tomato
        
    def test_remote(self):
        user= User(['cook'])
        policy = self.makeOne()
        tomato=self.make_nested_tomato()
        allowed=policy.has_permission(tomato, 'cook', attr='cookie.hotness', user=user)
        self.failUnless(not allowed)
        user= User(['cook', 'join'])
        allowed=policy.has_permission(tomato, 'cook', attr='chili.hotness', user=user)
        self.failUnless(allowed)
        allowed=policy.has_permission(tomato, 'cook', attr='chili2.hotness', user=user)
        self.failUnless(not allowed)
    
    def test_remote_nested(self):
        user= User(['cook', 'join'])
        policy=TestingPolicy()
        class ExtendedTestingPolicy(TestingPolicy):
            pass
        ExtendedTestingPolicy.register(
            predicate=any_permission('join'),
            obj=Chili,
            action='join')
        policy2=ExtendedTestingPolicy()
        tomato= self.make_nested_tomato()
        allowed=policy.has_permission(tomato.chili.cookie, 'cook', attr='sweetness', user=user)
        self.failUnless(allowed)
        allowed=policy.has_permission(tomato.chili, 'cook', attr='cookie.sweetness', user=user)
        self.failUnless(not allowed)
        allowed=policy.has_permission(tomato, 'cook', attr='chili.cookie.sweetness', user=user)
        self.failUnless(not allowed)
        allowed=policy2.has_permission(tomato, 'cook', attr='chili.cookie', user=user)
        self.failUnless(allowed)
        allowed=policy2.has_permission(tomato, 'cook', attr='chili.cookie.sweetness', user=user)
        self.failUnless(allowed)
        
    def test_attr_permissions(self):
        policy = self.makeOne()
        u=User(["cook"])
        allowed = policy.has_permission(Chili(),"cook", user=u)
        self.failUnless(allowed)
        allowed = policy.has_permission(Chili(),"cook", attr="hotness", user=u)
        self.failUnless(allowed)
        allowed = policy.has_permission(Chili(),"cook", attr="amount", user=u)
        self.failUnless(not allowed)